/*
 *  This file is part of the KDE libraries
 *  Copyright (c) 2001 Michael Goffioul <kdeprint@swing.be>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License version 2 as published by the Free Software Foundation.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public License
 *  along with this library; see the file COPYING.LIB.  If not, write to
 *  the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 **/

#include "cupsaddsmb2.h"
#include "cupsinfos.h"
#include "sidepixmap.h"

#include <QtCore/QTimer>
#include <QtGui/QProgressBar>
#include <QtGui/QLabel>
#include <QtGui/QLayout>
#include <klineedit.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <QtGui/QMessageBox>
#include <QtCore/QFile>
#include <kdebug.h>
#include <kseparator.h>
#include <kpushbutton.h>
#include <kstandardguiitem.h>

#include <cups/cups.h>
#include <ctype.h>

CupsAddSmb::CupsAddSmb(QWidget *parent, const char *name)
: KDialog(parent)
{
  setObjectName( name );
	m_state = None;
	m_status = false;
	m_actionindex = 0;
        m_proc.setOutputChannelMode(KProcess::MergedChannels);
	connect(&m_proc, SIGNAL(readyReadStandardOutput()), SLOT(slotReceived()));
	connect(&m_proc, SIGNAL(finished(int,KProcess::ExitStatus)),
                SLOT(slotProcessExited(int,KProcess::ExitStatus)));

	m_side = new SidePixmap(this);
	m_doit = new QPushButton(i18n("&Export"), this);
	m_cancel = new KPushButton(KStandardGuiItem::cancel(), this);
	connect(m_cancel, SIGNAL(clicked()), SLOT(reject()));
	connect(m_doit, SIGNAL(clicked()), SLOT(slotActionClicked()));
	m_bar = new QProgressBar(this);
	m_text = new QLabel(this);
        m_text->setOpenExternalLinks(true);
        m_text->setTextInteractionFlags(Qt::LinksAccessibleByMouse);
        QLabel	*m_title = new QLabel(i18n("Export Printer Driver to Windows Clients"), this);
	setCaption(m_title->text());
	QFont	f(m_title->font());
	f.setBold(true);
	m_title->setFont(f);
	KSeparator	*m_sep = new KSeparator(Qt::Horizontal, this);
	m_textinfo = new QLabel( this );
	m_logined = new KLineEdit( this );
	m_passwded = new KLineEdit( this );
	m_passwded->setPasswordMode( true );
	m_servered = new KLineEdit( this );
	QLabel *m_loginlab = new QLabel( i18n( "&Username:" ), this );
	QLabel *m_serverlab = new QLabel( i18n( "&Samba server:" ), this );
	QLabel *m_passwdlab = new QLabel( i18n( "&Password:" ), this );
	m_loginlab->setBuddy( m_logined );
	m_serverlab->setBuddy( m_servered );
	m_passwdlab->setBuddy( m_passwded );

	QString txt = i18n( "<p><b>Samba server</b></p>"
						"Adobe Windows PostScript driver files plus the CUPS printer PPD will be "
						"exported to the <tt>[print$]</tt> special share of the Samba server (to change "
						"the source CUPS server, use the <nobr><i>Configure Manager -> CUPS server</i></nobr> first). "
						"The <tt>[print$]</tt> share must exist on the Samba side prior to clicking the "
						"<b>Export</b> button below." );
	m_serverlab->setWhatsThis(txt );
	m_servered->setWhatsThis(txt );

	txt = i18n( "<p><b>Samba username</b></p>"
				"User needs to have write access to the <tt>[print$]</tt> share on the Samba server. "
				"<tt>[print$]</tt> holds printer drivers prepared for download to Windows clients. "
				"This dialog does not work for Samba servers configured with <tt>security = share</tt> "
				"(but works fine with <tt>security = user</tt>)." );
	m_loginlab->setWhatsThis(txt );
	m_logined->setWhatsThis(txt );

	txt = i18n( "<p><b>Samba password</b></p>"
				"The Samba setting <tt>encrypt passwords = yes</tt> "
				"(default) requires prior use of <tt>smbpasswd -a [username]</tt> command, "
				"to create an encrypted Samba password and have Samba recognize it." );
	m_passwdlab->setWhatsThis(txt );
	m_passwded->setWhatsThis(txt );

	QHBoxLayout	*l0 = new QHBoxLayout(this);
	l0->setMargin(10);
	l0->setSpacing(10);
	QVBoxLayout	*l1 = new QVBoxLayout();
	l1->setMargin(0);
	l1->setSpacing(10);
	l0->addWidget(m_side);
	l0->addLayout(l1);
	l1->addWidget(m_title);
	l1->addWidget(m_sep);
	l1->addWidget(m_text);
	QGridLayout *l3 = new QGridLayout();
    l3->setMargin(0);
    l3->setSpacing(10);
	l1->addLayout( l3 );
	l3->addWidget( m_loginlab, 1, 0 );
	l3->addWidget( m_passwdlab, 2, 0 );
	l3->addWidget( m_serverlab, 0, 0 );
	l3->addWidget( m_logined, 1, 1 );
	l3->addWidget( m_passwded, 2, 1 );
	l3->addWidget( m_servered, 0, 1 );
	l3->setColumnStretch( 1, 1 );
	l1->addSpacing( 10 );
	l1->addWidget(m_bar);
	l1->addWidget( m_textinfo );
	l1->addSpacing(30);
	QHBoxLayout	*l2 = new QHBoxLayout();
	l2->setMargin(0);
	l2->setSpacing(10);
	l1->addLayout(l2);
	l2->addStretch(1);
	l2->addWidget(m_doit);
	l2->addWidget(m_cancel);

	m_logined->setText( CupsInfos::self()->login() );
	m_passwded->setText( CupsInfos::self()->password() );
	m_servered->setText( cupsServer() );

	setMinimumHeight(400);
}

CupsAddSmb::~CupsAddSmb()
{
}

void CupsAddSmb::slotActionClicked()
{
	if (m_state == None)
		doExport();
	else if (m_proc.state() != KProcess::NotRunning)
		m_proc.kill();
}

void CupsAddSmb::slotReceived()
{
	QString	line;
	bool	partial(false);
	static bool incomplete(false);

	kDebug(500) << "slotReceived()";
	while (1)
	{
		// read a line
		line = QLatin1String("");
		partial = true;
		char c;
		while (m_proc.getChar(&c))
		{
			if (c == '\n')
			{
				partial = false;
				break;
			}
			else
			{
				QChar qc(c);
				if (qc.isPrint())
					line += qc;
			}
		}

		if (line.isEmpty())
		{
			kDebug(500) << "NOTHING TO READ";
			return;
		}

		kDebug(500) << "ANSWER = " << line << " (END = " << line.length() << ")";
		if (!partial)
		{
			if (incomplete && m_buffer.count() > 0)
				m_buffer[m_buffer.size()-1].append(line);
			else
				m_buffer << line;
			incomplete = false;
			kDebug(500) << "COMPLETE LINE READ (" << m_buffer.count() << ")";
		}
		else
		{
			if (line.startsWith("smb:") || line.startsWith("rpcclient $"))
			{
				kDebug(500) << "END OF ACTION";
				checkActionStatus();
				if (m_status)
					nextAction();
				else
				{
					// quit program
					kDebug(500) << "EXITING PROGRAM...";
					m_proc.write("quit\n", 5);
					kDebug(500) << "SENT";
				}
				return;
			}
			else
			{
				if (incomplete && m_buffer.count() > 0)
					m_buffer[m_buffer.size()-1].append(line);
				else
					m_buffer << line;
				incomplete = true;
				kDebug(500) << "INCOMPLETE LINE READ (" << m_buffer.count() << ")";
			}
		}
	}
}

void CupsAddSmb::checkActionStatus()
{
	m_status = false;
	// when checking the status, we need to take into account the
	// echo of the command in the output buffer.
	switch (m_state)
	{
		case None:
		case Start:
			m_status = true;
			break;
		case Copy:
			m_status = (m_buffer.count() == 0);
			break;
		case MkDir:
			m_status = (m_buffer.count() == 1 || m_buffer[1].indexOf("ERRfilexists") != -1);
			break;
		case AddDriver:
		case AddPrinter:
			m_status = (m_buffer.count() == 1 || !m_buffer[1].startsWith("result"));
			break;
	}
	kDebug(500) << "ACTION STATUS = " << m_status;
}

void CupsAddSmb::nextAction()
{
	if (m_actionindex < (int)(m_actions.count()))
		QTimer::singleShot(1, this, SLOT(doNextAction()));
}

void CupsAddSmb::doNextAction()
{
	m_buffer.clear();
	m_state = None;
	if (m_proc.state() == KProcess::Running)
	{
		QByteArray	s = m_actions[m_actionindex++].toLatin1();
		m_bar->setValue(m_bar->value()+1);
		kDebug(500) << "NEXT ACTION = " << s;
		if (s == "quit")
		{
			// do nothing
		}
		else if (s == "mkdir")
		{
			m_state = MkDir;
			//m_text->setText(i18n("Creating directory %1").arg(m_actions[m_actionindex]));
			m_textinfo->setText(i18n("Creating folder %1", m_actions[m_actionindex]));
			s.append(" ").append(m_actions[m_actionindex].toLatin1());
			m_actionindex++;
		}
		else if (s == "put")
		{
			m_state = Copy;
			//m_text->setText(i18n("Uploading %1").arg(m_actions[m_actionindex+1]));
			m_textinfo->setText(i18n("Uploading %1", m_actions[m_actionindex+1]));
			s.append(" ").append(QFile::encodeName(m_actions[m_actionindex]).data()).append(" ").append(m_actions[m_actionindex+1].toLatin1());
			m_actionindex += 2;
		}
		else if (s == "adddriver")
		{
			m_state = AddDriver;
			//m_text->setText(i18n("Installing driver for %1").arg(m_actions[m_actionindex]));
			m_textinfo->setText(i18n("Installing driver for %1", m_actions[m_actionindex]));
			s.append(" \"").append(m_actions[m_actionindex].toLatin1()).append("\" \"").append(m_actions[m_actionindex+1].toLatin1()).append("\"");
			m_actionindex += 2;
		}
		else if (s == "addprinter" || s == "setdriver")
		{
			m_state = AddPrinter;
			//m_text->setText(i18n("Installing printer %1").arg(m_actions[m_actionindex]));
			m_textinfo->setText(i18n("Installing printer %1", m_actions[m_actionindex]));
			QByteArray	dest = m_actions[m_actionindex].toLocal8Bit();
			if (s == "addprinter")
				s.append(" ").append(dest).append(" ").append(dest).append(" \"").append(dest).append("\" \"\"");
			else
				s.append(" ").append(dest).append(" ").append(dest);
			m_actionindex++;
		}
		else
		{
			kDebug(500) << "ACTION = unknown action";
			m_proc.kill();
			return;
		}
		// send action
		kDebug(500) << "ACTION = " << s;
		s.append("\n");
		m_proc.write(s);
	}
}

void CupsAddSmb::slotProcessExited(int, KProcess::ExitStatus exitStatus)
{
	kDebug(500) << "PROCESS EXITED (" << m_state << ")";
	if (exitStatus == KProcess::NormalExit && m_state != Start && m_status)
	{
		// last process went OK. If it was smbclient, then switch to rpcclient
		if (m_procname == "smbclient")
		{
			doInstall();
			return;
		}
		else
		{
			m_doit->setEnabled(false);
			m_cancel->setEnabled(true);
			m_cancel->setText(i18n("&Close"));
			m_cancel->setDefault(true);
			m_cancel->setFocus();
			m_logined->setEnabled( true );
			m_servered->setEnabled( true );
			m_passwded->setEnabled( true );
			m_text->setText(i18n("Driver successfully exported."));
			m_bar->reset();
			m_textinfo->setText( QString() );
			return;
		}
	}

	if (exitStatus == KProcess::NormalExit)
	{
		showError(
				i18n("Operation failed. Possible reasons are: permission denied "
				     "or invalid Samba configuration (see <a href=\"man:/cupsaddsmb\">"
					 "cupsaddsmb</a> manual page for detailed information, you need "
					 "<a href=\"http://www.cups.org\">CUPS</a> version 1.1.11 or higher). "
					 "You may want to try again with another login/password."));

	}
	else
	{
		showError(i18n("Operation aborted (process killed)."));
	}
}

void CupsAddSmb::showError(const QString& msg)
{
	m_text->setText(i18n("<h3>Operation failed.</h3><p>%1</p>", msg));
	m_cancel->setEnabled(true);
	m_logined->setEnabled( true );
	m_servered->setEnabled( true );
	m_passwded->setEnabled( true );
	m_doit->setText(i18n("&Export"));
	m_state = None;
}

bool CupsAddSmb::exportDest(const QString &dest, const QString& datadir)
{
	CupsAddSmb	dlg;
	dlg.m_dest = dest;
	dlg.m_datadir = datadir;
	dlg.m_text->setText(
			i18n( "You are about to prepare the <b>%1</b> driver to be "
			      "shared out to Windows clients through Samba. This operation "
				  "requires the <a href=\"http://www.adobe.com/products/printerdrivers/main.html\">Adobe PostScript Driver</a>, a recent version of "
				  "Samba 2.2.x and a running SMB service on the target server. "
				  "Click <b>Export</b> to start the operation. Read the <a href=\"man:/cupsaddsmb\">cupsaddsmb</a> "
				  "manual page in Konqueror or type <tt>man cupsaddsmb</tt> in a "
				  "console window to learn more about this functionality." ,  dest ) );
	dlg.exec();
	return dlg.m_status;
}

bool CupsAddSmb::doExport()
{
	m_status = false;
	m_state = None;

	if (!QFile::exists(m_datadir+"/drivers/ADOBEPS5.DLL") ||
	    !QFile::exists(m_datadir+"/drivers/ADOBEPS4.DRV"))
	{
		showError(
			i18n("Some driver files are missing. You can get them on "
			     "<a href=\"http://www.adobe.com\">Adobe</a> web site. "
			     "See <a href=\"man:/cupsaddsmb\">cupsaddsmb</a> manual "
				 "page for more details (you need <a href=\"http://www.cups.org\">CUPS</a> "
				 "version 1.1.11 or higher)."));
		return false;
	}

	m_bar->setRange(0, 18);
	m_bar->setValue(0);
	//m_text->setText(i18n("<p>Preparing to upload driver to host <b>%1</b>").arg(m_servered->text()));
	m_textinfo->setText(i18n("Preparing to upload driver to host %1", m_servered->text()));
	m_cancel->setEnabled(false);
	m_logined->setEnabled( false );
	m_servered->setEnabled( false );
	m_passwded->setEnabled( false );
	m_doit->setText(i18n("&Abort"));

	const char	*ppdfile;

	if ((ppdfile = cupsGetPPD(m_dest.toLocal8Bit())) == NULL)
	{
		showError(i18n("The driver for printer <b>%1</b> could not be found.", m_dest));
		return false;
	}

	m_actions.clear();
	m_actions << "mkdir" << "W32X86";
	m_actions << "put" << ppdfile << "W32X86/"+m_dest+".PPD";
	m_actions << "put" << m_datadir+"/drivers/ADOBEPS5.DLL" << "W32X86/ADOBEPS5.DLL";
	m_actions << "put" << m_datadir+"/drivers/ADOBEPSU.DLL" << "W32X86/ADOBEPSU.DLL";
	m_actions << "put" << m_datadir+"/drivers/ADOBEPSU.HLP" << "W32X86/ADOBEPSU.HLP";
	m_actions << "mkdir" << "WIN40";
	m_actions << "put" << ppdfile << "WIN40/"+m_dest+".PPD";
	m_actions << "put" << m_datadir+"/drivers/ADFONTS.MFM" << "WIN40/ADFONTS.MFM";
	m_actions << "put" << m_datadir+"/drivers/ADOBEPS4.DRV" << "WIN40/ADOBEPS4.DRV";
	m_actions << "put" << m_datadir+"/drivers/ADOBEPS4.HLP" << "WIN40/ADOBEPS4.HLP";
	m_actions << "put" << m_datadir+"/drivers/DEFPRTR2.PPD" << "WIN40/DEFPRTR2.PPD";
	m_actions << "put" << m_datadir+"/drivers/ICONLIB.DLL" << "WIN40/ICONLIB.DLL";
	m_actions << "put" << m_datadir+"/drivers/PSMON.DLL" << "WIN40/PSMON.DLL";
	m_actions << "quit";

	m_proc.clearProgram();
	m_procname = "smbclient";
	m_proc << m_procname << QLatin1String("//")+m_servered->text()+"/print$";
	return startProcess();
}

bool CupsAddSmb::doInstall()
{
	m_status = false;
	m_state = None;

	m_actions.clear();
	m_actions << "adddriver" << "Windows NT x86" << m_dest+":ADOBEPS5.DLL:"+m_dest+".PPD:ADOBEPSU.DLL:ADOBEPSU.HLP:NULL:RAW:NULL";
	// seems to be wrong with newer versions of Samba
	//m_actions << "addprinter" << m_dest;
	m_actions << "adddriver" << "Windows 4.0" << m_dest+":ADOBEPS4.DRV:"+m_dest+".PPD:NULL:ADOBEPS4.HLP:PSMON.DLL:RAW:ADFONTS.MFM,DEFPRTR2.PPD,ICONLIB.DLL";
	// seems to be ok with newer versions of Samba
	m_actions << "setdriver" << m_dest;
	m_actions << "quit";

	//m_text->setText(i18n("Preparing to install driver on host <b>%1</b>").arg(m_servered->text()));
	m_textinfo->setText(i18n("Preparing to install driver on host %1", m_servered->text()));

	m_proc.clearProgram();
	m_procname = "rpcclient";
	m_proc << m_procname << m_servered->text();
	return startProcess();
}

bool CupsAddSmb::startProcess()
{
	m_proc << "-d" << "0" << "-N" << "-U";
	if (m_passwded->text().isEmpty())
		m_proc << m_logined->text();
	else
		m_proc << m_logined->text()+'%'+m_passwded->text();
	m_state = Start;
	m_actionindex = 0;
	m_buffer.clear();
	kDebug(500) << "PROCESS STARTED = " << m_procname;
	m_proc.start();
	return m_proc.waitForStarted(5000); // don't wait more than 5 seconds
}

#include "cupsaddsmb2.moc"
